﻿// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.
// See the LICENSE file in the project root for more information.

using System;
using System.Collections.Generic;
using System.Collections.Immutable;
using System.Diagnostics;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using Microsoft.CodeAnalysis.FindSymbols;
using Microsoft.CodeAnalysis.PatternMatching;
using Microsoft.CodeAnalysis.PersistentStorage;
using Microsoft.CodeAnalysis.PooledObjects;
using Microsoft.CodeAnalysis.Shared.Extensions;
using Microsoft.CodeAnalysis.Text;
using Roslyn.Utilities;

namespace Microsoft.CodeAnalysis.NavigateTo
{
    internal abstract partial class AbstractNavigateToSearchService
    {
        private static ImmutableArray<(PatternMatchKind roslynKind, NavigateToMatchKind vsKind)> s_kindPairs =
            ImmutableArray.Create(
                (PatternMatchKind.Exact, NavigateToMatchKind.Exact),
                (PatternMatchKind.Prefix, NavigateToMatchKind.Prefix),
                (PatternMatchKind.NonLowercaseSubstring, NavigateToMatchKind.Substring),
                (PatternMatchKind.StartOfWordSubstring, NavigateToMatchKind.Substring),
                (PatternMatchKind.CamelCaseExact, NavigateToMatchKind.CamelCaseExact),
                (PatternMatchKind.CamelCasePrefix, NavigateToMatchKind.CamelCasePrefix),
                (PatternMatchKind.CamelCaseNonContiguousPrefix, NavigateToMatchKind.CamelCaseNonContiguousPrefix),
                (PatternMatchKind.CamelCaseSubstring, NavigateToMatchKind.CamelCaseSubstring),
                (PatternMatchKind.CamelCaseNonContiguousSubstring, NavigateToMatchKind.CamelCaseNonContiguousSubstring),
                (PatternMatchKind.Fuzzy, NavigateToMatchKind.Fuzzy),
                // Map our value to 'Fuzzy' as that's the lower value the platform supports.
                (PatternMatchKind.LowercaseSubstring, NavigateToMatchKind.Fuzzy));

        public static Task SearchFullyLoadedProjectInCurrentProcessAsync(
            Project project, ImmutableArray<Document> priorityDocuments, string searchPattern,
            IImmutableSet<string> kinds, Func<RoslynNavigateToItem, Task> onResultFound, CancellationToken cancellationToken)
        {
            return FindSearchResultsAsync(
                project, priorityDocuments, searchDocument: null, pattern: searchPattern, kinds, onResultFound, cancellationToken);
        }

        public static Task SearchFullyLoadedDocumentInCurrentProcessAsync(
            Document document, string searchPattern, IImmutableSet<string> kinds,
            Func<RoslynNavigateToItem, Task> onResultFound, CancellationToken cancellationToken)
        {
            return FindSearchResultsAsync(
                document.Project, priorityDocuments: ImmutableArray<Document>.Empty,
                document, searchPattern, kinds, onResultFound, cancellationToken);
        }

        private static async Task FindSearchResultsAsync(
            Project project, ImmutableArray<Document> priorityDocuments,
            Document? searchDocument, string pattern, IImmutableSet<string> kinds,
            Func<RoslynNavigateToItem, Task> onResultFound, CancellationToken cancellationToken)
        {
            // If the user created a dotted pattern then we'll grab the last part of the name
            var (patternName, patternContainerOpt) = PatternMatcher.GetNameAndContainer(pattern);

            var declaredSymbolInfoKindsSet = new DeclaredSymbolInfoKindSet(kinds);

            // Prioritize the active documents if we have any.
            var highPriDocs = priorityDocuments.Where(d => project.ContainsDocument(d.Id)).ToSet();
            await ProcessDocumentsAsync(searchDocument, patternName, patternContainerOpt, declaredSymbolInfoKindsSet, onResultFound, highPriDocs, cancellationToken).ConfigureAwait(false);

            // Then process non-priority documents.
            var lowPriDocs = project.Documents.Where(d => !highPriDocs.Contains(d)).ToSet();
            await ProcessDocumentsAsync(searchDocument, patternName, patternContainerOpt, declaredSymbolInfoKindsSet, onResultFound, lowPriDocs, cancellationToken).ConfigureAwait(false);

            // if the caller is only searching a single doc, and we already covered it above, don't bother computing
            // source-generator docs.
            if (searchDocument != null && (highPriDocs.Contains(searchDocument) || lowPriDocs.Contains(searchDocument)))
                return;

            // Finally, generate and process and source-generated docs.  this may take some time, so we always want to
            // do this after the other documents.
            var generatedDocs = await project.GetSourceGeneratedDocumentsAsync(cancellationToken).ConfigureAwait(false);
            await ProcessDocumentsAsync(searchDocument, patternName, patternContainerOpt, declaredSymbolInfoKindsSet, onResultFound, generatedDocs.ToSet<Document>(), cancellationToken).ConfigureAwait(false);
        }

        private static async Task ProcessDocumentsAsync(
            Document? searchDocument, string patternName, string? patternContainer, DeclaredSymbolInfoKindSet kinds,
            Func<RoslynNavigateToItem, Task> onResultFound, ISet<Document> documents, CancellationToken cancellationToken)
        {
            using var _ = ArrayBuilder<Task>.GetInstance(out var tasks);

            foreach (var document in documents)
            {
                if (searchDocument != null && searchDocument != document)
                    continue;

                cancellationToken.ThrowIfCancellationRequested();
                tasks.Add(Task.Run(() =>
                    ProcessDocumentAsync(document, patternName, patternContainer, kinds, onResultFound, cancellationToken), cancellationToken));
            }

            await Task.WhenAll(tasks).ConfigureAwait(false);
        }

        private static async Task ProcessDocumentAsync(
            Document document, string patternName, string? patternContainer, DeclaredSymbolInfoKindSet kinds,
            Func<RoslynNavigateToItem, Task> onResultFound, CancellationToken cancellationToken)
        {
            var index = await SyntaxTreeIndex.GetRequiredIndexAsync(document, cancellationToken).ConfigureAwait(false);

            await ProcessIndexAsync(
                document.Id, document, patternName, patternContainer, kinds, onResultFound, index, cancellationToken).ConfigureAwait(false);
        }

        public static async Task SearchCachedDocumentsInCurrentProcessAsync(
            Workspace workspace,
            ImmutableArray<DocumentKey> documentKeys,
            ImmutableArray<DocumentKey> priorityDocumentKeys,
            string searchPattern,
            IImmutableSet<string> kinds,
            Func<RoslynNavigateToItem, Task> onItemFound,
            CancellationToken cancellationToken)
        {
            var stringTable = new StringTable();

            var highPriDocsSet = priorityDocumentKeys.ToSet();
            var lowPriDocs = documentKeys.WhereAsArray(d => !highPriDocsSet.Contains(d));

            // If the user created a dotted pattern then we'll grab the last part of the name
            var (patternName, patternContainer) = PatternMatcher.GetNameAndContainer(searchPattern);
            var declaredSymbolInfoKindsSet = new DeclaredSymbolInfoKindSet(kinds);

            await SearchCachedDocumentsInCurrentProcessAsync(
                workspace, priorityDocumentKeys, patternName, patternContainer, declaredSymbolInfoKindsSet, onItemFound, stringTable, cancellationToken).ConfigureAwait(false);

            await SearchCachedDocumentsInCurrentProcessAsync(
                workspace, lowPriDocs, patternName, patternContainer, declaredSymbolInfoKindsSet, onItemFound, stringTable, cancellationToken).ConfigureAwait(false);
        }

        private static async Task SearchCachedDocumentsInCurrentProcessAsync(
            Workspace workspace,
            ImmutableArray<DocumentKey> documentKeys,
            string patternName,
            string patternContainer,
            DeclaredSymbolInfoKindSet kinds,
            Func<RoslynNavigateToItem, Task> onItemFound,
            StringTable stringTable,
            CancellationToken cancellationToken)
        {
            using var _ = ArrayBuilder<Task>.GetInstance(out var tasks);

            foreach (var documentKey in documentKeys)
            {
                tasks.Add(Task.Run(async () =>
                {
                    var index = await SyntaxTreeIndex.LoadAsync(
                        workspace, documentKey, checksum: null, stringTable, cancellationToken).ConfigureAwait(false);
                    if (index == null)
                        return;

                    await ProcessIndexAsync(
                        documentKey.Id, document: null, patternName, patternContainer, kinds, onItemFound, index, cancellationToken).ConfigureAwait(false);
                }, cancellationToken));
            }

            await Task.WhenAll(tasks).ConfigureAwait(false);
        }

        private static async Task ProcessIndexAsync(
            DocumentId documentId, Document? document,
            string patternName, string? patternContainer,
            DeclaredSymbolInfoKindSet kinds,
            Func<RoslynNavigateToItem, Task> onResultFound,
            SyntaxTreeIndex index,
            CancellationToken cancellationToken)
        {
            var containerMatcher = patternContainer != null
                ? PatternMatcher.CreateDotSeparatedContainerMatcher(patternContainer)
                : null;

            using var nameMatcher = PatternMatcher.CreatePatternMatcher(patternName, includeMatchedSpans: true, allowFuzzyMatching: true);
            using var _1 = containerMatcher;
            using var _2 = ArrayBuilder<PatternMatch>.GetInstance(out var nameMatches);
            using var _3 = ArrayBuilder<PatternMatch>.GetInstance(out var containerMatches);

            foreach (var declaredSymbolInfo in index.DeclaredSymbolInfos)
            {
                await AddResultIfMatchAsync(
                    documentId, document,
                    declaredSymbolInfo,
                    nameMatcher, containerMatcher,
                    kinds,
                    nameMatches, containerMatches,
                    onResultFound, cancellationToken).ConfigureAwait(false);
            }
        }

        private static async Task AddResultIfMatchAsync(
            DocumentId documentId, Document? document,
            DeclaredSymbolInfo declaredSymbolInfo,
            PatternMatcher nameMatcher, PatternMatcher? containerMatcher,
            DeclaredSymbolInfoKindSet kinds,
            ArrayBuilder<PatternMatch> nameMatches, ArrayBuilder<PatternMatch> containerMatches,
            Func<RoslynNavigateToItem, Task> onResultFound, CancellationToken cancellationToken)
        {
            nameMatches.Clear();
            containerMatches.Clear();

            cancellationToken.ThrowIfCancellationRequested();
            if (kinds.Contains(declaredSymbolInfo.Kind) &&
                nameMatcher.AddMatches(declaredSymbolInfo.Name, nameMatches) &&
                containerMatcher?.AddMatches(declaredSymbolInfo.FullyQualifiedContainerName, containerMatches) != false)
            {
                var result = await ConvertResultAsync(
                    documentId, document, declaredSymbolInfo, nameMatches, containerMatches, cancellationToken).ConfigureAwait(false);
                await onResultFound(result).ConfigureAwait(false);
            }
        }

        private static async Task<RoslynNavigateToItem> ConvertResultAsync(
            DocumentId documentId,
            Document? document,
            DeclaredSymbolInfo declaredSymbolInfo,
            ArrayBuilder<PatternMatch> nameMatches,
            ArrayBuilder<PatternMatch> containerMatches,
            CancellationToken cancellationToken)
        {
            var matchKind = GetNavigateToMatchKind(nameMatches);

            // A match is considered to be case sensitive if all its constituent pattern matches are
            // case sensitive. 
            var isCaseSensitive = nameMatches.All(m => m.IsCaseSensitive) && containerMatches.All(m => m.IsCaseSensitive);
            var kind = GetItemKind(declaredSymbolInfo);

            using var _ = ArrayBuilder<TextSpan>.GetInstance(out var matchedSpans);
            foreach (var match in nameMatches)
                matchedSpans.AddRange(match.MatchedSpans);

            // See if we have a match in a linked file.  If so, see if we have the same match in
            // other projects that this file is linked in.  If so, include the full set of projects
            // the match is in so we can display that well in the UI.
            //
            // We can only do this in the case where the solution is loaded and thus we can examine
            // the relationship between this document and the other documents linked to it.  In the
            // case where the solution isn't fully loaded and we're just reading in cached data, we
            // don't know what other files we're linked to and can't merge results in this fashion.
            var additionalMatchingProjects = document == null
                ? ImmutableArray<ProjectId>.Empty
                : await GetAdditionalProjectsWithMatchAsync(document, declaredSymbolInfo, cancellationToken).ConfigureAwait(false);

            // If we were not given a Document instance, then we're finding matches in cached data
            // and thus could be 'stale'.
            return new RoslynNavigateToItem(
                isStale: document == null,
                documentId,
                additionalMatchingProjects,
                declaredSymbolInfo,
                kind,
                matchKind,
                isCaseSensitive,
                matchedSpans.ToImmutable());
        }

        private static async Task<ImmutableArray<ProjectId>> GetAdditionalProjectsWithMatchAsync(
            Document document, DeclaredSymbolInfo declaredSymbolInfo, CancellationToken cancellationToken)
        {
            using var _ = ArrayBuilder<ProjectId>.GetInstance(out var result);

            var solution = document.Project.Solution;
            var linkedDocumentIds = document.GetLinkedDocumentIds();
            foreach (var linkedDocumentId in linkedDocumentIds)
            {
                var linkedDocument = solution.GetRequiredDocument(linkedDocumentId);
                var index = await SyntaxTreeIndex.GetRequiredIndexAsync(linkedDocument, cancellationToken).ConfigureAwait(false);

                // See if the index for the other file also contains this same info.  If so, merge the results so the
                // user only sees them as a single hit in the UI.
                if (index.DeclaredSymbolInfoSet.Contains(declaredSymbolInfo))
                    result.Add(linkedDocument.Project.Id);
            }

            result.RemoveDuplicates();
            return result.ToImmutable();
        }

        private static string GetItemKind(DeclaredSymbolInfo declaredSymbolInfo)
        {
            switch (declaredSymbolInfo.Kind)
            {
                case DeclaredSymbolInfoKind.Class:
                case DeclaredSymbolInfoKind.Record:
                    return NavigateToItemKind.Class;
                case DeclaredSymbolInfoKind.RecordStruct:
                    return NavigateToItemKind.Structure;
                case DeclaredSymbolInfoKind.Constant:
                    return NavigateToItemKind.Constant;
                case DeclaredSymbolInfoKind.Delegate:
                    return NavigateToItemKind.Delegate;
                case DeclaredSymbolInfoKind.Enum:
                    return NavigateToItemKind.Enum;
                case DeclaredSymbolInfoKind.EnumMember:
                    return NavigateToItemKind.EnumItem;
                case DeclaredSymbolInfoKind.Event:
                    return NavigateToItemKind.Event;
                case DeclaredSymbolInfoKind.Field:
                    return NavigateToItemKind.Field;
                case DeclaredSymbolInfoKind.Interface:
                    return NavigateToItemKind.Interface;
                case DeclaredSymbolInfoKind.Constructor:
                case DeclaredSymbolInfoKind.ExtensionMethod:
                case DeclaredSymbolInfoKind.Method:
                    return NavigateToItemKind.Method;
                case DeclaredSymbolInfoKind.Module:
                    return NavigateToItemKind.Module;
                case DeclaredSymbolInfoKind.Indexer:
                case DeclaredSymbolInfoKind.Property:
                    return NavigateToItemKind.Property;
                case DeclaredSymbolInfoKind.Struct:
                    return NavigateToItemKind.Structure;
                default:
                    throw ExceptionUtilities.UnexpectedValue(declaredSymbolInfo.Kind);
            }
        }

        private static NavigateToMatchKind GetNavigateToMatchKind(ArrayBuilder<PatternMatch> nameMatches)
        {
            // work backwards through the match kinds.  That way our result is as bad as our worst match part.  For
            // example, say the user searches for `Console.Write` and we find `Console.Write` (exact, exact), and
            // `Console.WriteLine` (exact, prefix).  We don't want the latter hit to be considered an `exact` match, and
            // thus as good as `Console.Write`.

            for (var i = s_kindPairs.Length - 1; i >= 0; i--)
            {
                var (roslynKind, vsKind) = s_kindPairs[i];
                foreach (var match in nameMatches)
                {
                    if (match.Kind == roslynKind)
                        return vsKind;
                }
            }

            return NavigateToMatchKind.Regular;
        }

        private readonly struct DeclaredSymbolInfoKindSet
        {
            private readonly ImmutableArray<bool> _lookupTable;

            public DeclaredSymbolInfoKindSet(IEnumerable<string> navigateToItemKinds)
            {
                // The 'Contains' method implementation assumes that the DeclaredSymbolInfoKind type is unsigned.
                Debug.Assert(Enum.GetUnderlyingType(typeof(DeclaredSymbolInfoKind)) == typeof(byte));

                var lookupTable = new bool[Enum.GetValues(typeof(DeclaredSymbolInfoKind)).Length];
                foreach (var navigateToItemKind in navigateToItemKinds)
                {
                    switch (navigateToItemKind)
                    {
                        case NavigateToItemKind.Class:
                            lookupTable[(int)DeclaredSymbolInfoKind.Class] = true;
                            lookupTable[(int)DeclaredSymbolInfoKind.Record] = true;
                            break;
                        case NavigateToItemKind.Constant:
                            lookupTable[(int)DeclaredSymbolInfoKind.Constant] = true;
                            break;

                        case NavigateToItemKind.Delegate:
                            lookupTable[(int)DeclaredSymbolInfoKind.Delegate] = true;
                            break;

                        case NavigateToItemKind.Enum:
                            lookupTable[(int)DeclaredSymbolInfoKind.Enum] = true;
                            break;

                        case NavigateToItemKind.EnumItem:
                            lookupTable[(int)DeclaredSymbolInfoKind.EnumMember] = true;
                            break;

                        case NavigateToItemKind.Event:
                            lookupTable[(int)DeclaredSymbolInfoKind.Event] = true;
                            break;

                        case NavigateToItemKind.Field:
                            lookupTable[(int)DeclaredSymbolInfoKind.Field] = true;
                            break;

                        case NavigateToItemKind.Interface:
                            lookupTable[(int)DeclaredSymbolInfoKind.Interface] = true;
                            break;

                        case NavigateToItemKind.Method:
                            lookupTable[(int)DeclaredSymbolInfoKind.Constructor] = true;
                            lookupTable[(int)DeclaredSymbolInfoKind.ExtensionMethod] = true;
                            lookupTable[(int)DeclaredSymbolInfoKind.Method] = true;
                            break;

                        case NavigateToItemKind.Module:
                            lookupTable[(int)DeclaredSymbolInfoKind.Module] = true;
                            break;

                        case NavigateToItemKind.Property:
                            lookupTable[(int)DeclaredSymbolInfoKind.Indexer] = true;
                            lookupTable[(int)DeclaredSymbolInfoKind.Property] = true;
                            break;

                        case NavigateToItemKind.Structure:
                            lookupTable[(int)DeclaredSymbolInfoKind.Struct] = true;
                            lookupTable[(int)DeclaredSymbolInfoKind.RecordStruct] = true;
                            break;

                        default:
                            // Not a recognized symbol info kind
                            break;
                    }
                }

                _lookupTable = ImmutableArray.CreateRange(lookupTable);
            }

            public bool Contains(DeclaredSymbolInfoKind item)
            {
                return (int)item < _lookupTable.Length
                    && _lookupTable[(int)item];
            }
        }
    }
}
